"""
https://www.trafiklab.se/api/trafiklab-apis/sl/deviations
"""

from typing import List, Optional, TypedDict, NotRequired

from .common import TransportMode

Publish = TypedDict(
    "Publish",
    {
        "from": str,
        "upto": Optional[str],
    },
)


class Priority(TypedDict):
    importance_level: int
    influence_level: int
    urgency_level: int


class MessageVariant(TypedDict):
    header: str
    details: str
    scope_alias: str
    language: str
    weblink: NotRequired[str]


class StopPoint(TypedDict):
    id: int
    name: str


class ScopeArea(TypedDict):
    id: int
    transport_authority: int
    name: str
    type: str
    stop_points: NotRequired[List[StopPoint]]


class Line(TypedDict):
    id: int
    transport_authority: int
    transport_mode: TransportMode
    designation: NotRequired[str]
    name: NotRequired[str]
    group_of_lines: NotRequired[str]


class Scope(TypedDict):
    """Entities affected by the deviation"""

    stop_areas: NotRequired[List[ScopeArea]]
    lines: NotRequired[List[Line]]


class Deviation(TypedDict):
    """information regarding deviations on SLs transport network."""

    version: int
    created: str
    publish: Publish
    priority: Priority
    message_variants: List[MessageVariant]
    modified: str
    deviation_case_id: NotRequired[int]
    scope: NotRequired[Scope]



from dataclasses import dataclass
from datetime import timedelta
from enum import StrEnum
from typing import Any, List, NotRequired, Tuple, TypedDict

from .common import CoordTuple, ProductClass, PropertiesType
from .stops import StopFinderType


class SearchType(StrEnum):
    # search by coordinates
    COORD = "coord"

    # search by street and stop names
    ANY = "any"


@dataclass(frozen=True)
class SearchLeg:
    type: SearchType

    value: str
    """
    location or coordinates.
    Coordinate are in format: '<x>:<y>:WGS84[dd.ddddd]'
    """

    @property
    def coordinates(self) -> Tuple[str, str]:
        """returns coordinates in format 'lat,lon'"""
        if self.type != SearchType.COORD:
            raise ValueError("SearchLeg is not of type COORD")
        coords = self.value.rsplit(":", maxsplit=1)[0]
        x, y = coords.split(":")
        return x, y

    @classmethod
    def from_coordinates(cls, lat: str, lon: str) -> "SearchLeg":
        """creates SearchLeg from coordinates"""
        return cls(SearchType.COORD, f"{lon}:{lat}:WGS84[dd.ddddd]")

    @classmethod
    def from_any(cls, value: str) -> "SearchLeg":
        """creates SearchLeg from street or stop id"""
        return cls(SearchType.ANY, value)

    @classmethod
    def from_stop_finder(cls, stop: StopFinderType) -> "SearchLeg":
        """creates SearchLeg from StopFinderType"""
        return cls(SearchType.ANY, stop["id"])

class Language(StrEnum):
    """Supported languages for the Journey Planner API."""

    SV = "sv"  # Swedish
    EN = "en"  # English


class RouteType(StrEnum):
    LEAST_INTER_CHANGES = "leastinterchange"
    LEAST_TIME = "leasttime"
    LEAST_WALKING = "leastwalking"


class DwellTime:
    """Represents dwell time in a journey."""

    def __init__(self, duration: timedelta):
        self.duration = duration

    def to_hh_mm(self) -> str:
        """returns dwell time in format 'hh:mm'"""
        total_seconds = int(self.duration.total_seconds())
        if total_seconds >= 24 * 3600:
            raise ValueError("Dwell time cannot exceed 24 hours")

        hours, remainder = divmod(total_seconds, 3600)
        minutes, _ = divmod(remainder, 60)
        return f"{hours:02}:{minutes:02}"


class Manouvre(StrEnum):
    ORIGIN = "ORIGIN"
    DESTINATION = "DESTINATION"
    KEEP = "KEEP"
    TURN = "TURN"


class TurnDirection(StrEnum):
    UNKNOWN = "UNKNOWN"
    STRAIGHT = "STRAIGHT"
    SHARP_LEFT = "SHARP_LEFT"
    RIGHT = "RIGHT"
    LEFT = "LEFT"


class LegPoint(TypedDict):
    coord: CoordTuple
    cumDistance: int
    cumDuration: int
    distance: int
    duration: int
    fromCoordsIndex: int
    manouvre: Manouvre
    properties: NotRequired[PropertiesType]
    name: str
    skyDirection: NotRequired[int]
    toCoordsIndex: int
    turnDirection: TurnDirection


class JourneyOrigin(StopFinderType):
    departureTimeBaseTimetable: NotRequired[str]
    departureTimePlanned: NotRequired[str]
    departureTimeEstimated: NotRequired[str]
    properties: NotRequired[PropertiesType]


class JourneyDestination(StopFinderType):
    arrivalTimeBaseTimetable: NotRequired[str]
    arrivalTimePlanned: NotRequired[str]
    arrivalTimeEstimated: NotRequired[str]
    properties: NotRequired[PropertiesType]


class JourneyStop(StopFinderType):
    departureTimePlanned: NotRequired[str]
    arrivalTimePlanned: NotRequired[str]


TransportationProduct = TypedDict(
    "TransportationProduct",
    {"id": NotRequired[int], "class": ProductClass, "name": str, "iconId": int},
)


class TransportationOperator(TypedDict):
    id: str
    name: str


class TransportationDestination(TypedDict):
    id: str
    name: str
    type: str


class JourneyTransportation(TypedDict):
    id: NotRequired[str]
    name: NotRequired[str]
    disassembledName: NotRequired[str]
    number: NotRequired[str]
    product: TransportationProduct
    operator: NotRequired[TransportationOperator]
    destination: NotRequired[TransportationDestination]
    properties: PropertiesType


class FootPathLeg(TypedDict):
    duration: int
    footPathElem: List[Any]
    position: str


class JoyrneyLeg(TypedDict):
    infos: List[Any]
    distance: NotRequired[int]
    duration: int
    footPathInfo: NotRequired[List[FootPathLeg]]
    pathDescriptions: NotRequired[List[LegPoint]]
    origin: JourneyOrigin
    destination: JourneyDestination
    transportation: JourneyTransportation
    stopSequence: List[JourneyStop]
    properties: PropertiesType
    coords: NotRequired[List[CoordTuple]]
    realtimeStatus: NotRequired[List[str]]
    isRealtimeControlled: NotRequired[bool]


class DaysOfService(TypedDict):
    rvb: str


class Journey(TypedDict):
    tripDuration: int
    tripRtDuration: int
    rating: int
    isAdditional: bool
    interchanges: int
    legs: List[JoyrneyLeg]
    daysOfService: DaysOfService



from enum import IntEnum, StrEnum
from typing import Any, Dict, Tuple, TypeAlias

PropertiesType: TypeAlias = Dict[str, Any]
CoordTuple: TypeAlias = Tuple[float, float]

class TransportMode(StrEnum):
    BUS = "BUS"
    METRO = "METRO"
    TRAM = "TRAM"
    TRAIN = "TRAIN"
    SHIP = "SHIP"
    FERRY = "FERRY"
    TAXI = "TAXI"

class ProductClass(IntEnum):
    TRAIN = 0
    METRO = 2
    LOCAL_TRAIN = 4
    TRAM = 4
    BUS = 5
    SHIP_AND_FERRY = 9
    TRANSIT_ON_DEMAND_AREA_SERVICE = 10
    LONG_DISTANCE_TRAIN = 14
    EXPRESS_TRAIN = 14

    FOOT_PATH = 100
    FOOT_PATH_LOCAL = 99



import pytest

from tsl.utils import global_id_to_site_id


@pytest.mark.parametrize(
    "global_id, expected_site_id",
    [
        ("9091001000005730", 5730),
        ("9091001000009731", 9731),
    ])
def test_global_id_to_site_id(global_id, expected_site_id):
    assert global_id_to_site_id(global_id) == expected_site_id



from importlib.metadata import PackageNotFoundError, version
from typing import TYPE_CHECKING

try:
    if TYPE_CHECKING:
        assert __package__

    __version__ = version(__package__)
except PackageNotFoundError:
    __version__ = "0.0.0+local"



from typing import Any, List, Optional, cast
from urllib.parse import quote

from ..models.departures import SiteDepartureResponse, TransportMode
from ..models.sites import Site
from .common import AsyncClient, UrlParams

__all__ = ("TransportClient",)


class TransportClient(AsyncClient):
    """
    client for SL Transport API
    https://www.trafiklab.se/api/trafiklab-apis/sl/transport/

    only departures and sites are supported at the moment
    """

    @staticmethod
    def get_departures_url_params(
        site_id: int,
        transport: Optional[TransportMode] = None,
        direction: Optional[int] = None,
        line: Optional[int] = None,
        forecast: int = 60,
    ) -> UrlParams:
        url = (
            "https://transport.integration.sl.se"
            f"/v1/sites/{quote(str(site_id))}/departures"
        )
        params: dict[str, Any] = {}
        if transport is not None:
            params["transport"] = transport.value
        if direction is not None:
            params["direction"] = direction
        if line is not None:
            params["line"] = line
        if forecast is not None:
            params["forecast"] = forecast

        return UrlParams(url, params)

    async def get_site_departures(
        self,
        site_id: int,
        transport: Optional[TransportMode] = None,
        direction: Optional[int] = None,
        line: Optional[int] = None,
        forecast: int = 60,
    ) -> SiteDepartureResponse:
        """
        Get upcoming departures and deviations starting from time of
        the request (a maximum of 3 departures for each line & direction)
        """

        args = self.get_departures_url_params(
            site_id, transport, direction, line, forecast
        )

        response = await self._request_json(args)
        return cast(SiteDepartureResponse, response)

    async def get_sites(self) -> List[Site]:
        """List all sites within Region Stockholm"""

        args = UrlParams("https://transport.integration.sl.se/v1/sites", None)
        response = await self._request_json(args)
        return cast(List[Site], response)



import aiohttp
import pytest
import sys

from tsl.clients.deviations import DeviationsClient
from tsl.clients.stoplookup import StopLookupClient
from tsl.clients.transport import TransportClient
from tsl.models.departures import DepartureState
from tsl.models.deviations import TransportMode


@pytest.fixture
async def session():
    async with aiohttp.ClientSession() as session:
        yield session


@pytest.mark.integration
async def test_deviations(session):
    cl = DeviationsClient(session)
    deviations = await cl.get_deviations(
        line=["10", "43"],
        transport_mode=[TransportMode.METRO, TransportMode.TRAIN],
    )

    assert isinstance(deviations, list)
    if deviation := next(iter(deviations), None):
        assert isinstance(deviation, dict)
        assert isinstance(deviation["version"], int)


@pytest.mark.integration
async def test_transport_departures(session):
    cl = TransportClient(session)
    response = await cl.get_site_departures(1002)

    if departure := next(iter(response["departures"]), None):
        if sys.version_info < (3, 12):
            assert departure["state"] in DepartureState._value2member_map_
        else:
            assert departure["state"] in DepartureState



@pytest.mark.integration
async def test_transport_sites(session):
    cl = TransportClient(session)
    sites = await cl.get_sites()

    assert isinstance(sites, list)

    if site := next(iter(sites), None):
        assert isinstance(site, dict)
        assert isinstance(site["id"], int)


@pytest.mark.integration
async def test_stop_lookup(session):
    cl = StopLookupClient(session)
    stops = await cl.get_stops("Odenplan")
    assert isinstance(stops, list)

    if stops:
        assert stops[0]["disassembledName"] == "Odenplan"
        assert stops[0]["id"] == "9091001000009117"



from enum import StrEnum
from typing import List, TypedDict, NotRequired

from .common import CoordTuple, ProductClass


class StopFinderResultType(StrEnum):
    ADDRESS = "address"
    STOP = "stop"
    PLATFORM = "platform"
    SINGLEHOUSE = "singlehouse"
    POI = "poi"
    STREET = "street"


class StopArea(TypedDict):
    id: str
    name: str
    type: str


class StopFinderType(TypedDict):
    # The id of the search result.
    id: str

    # Value is true if location is a stop.
    isGlobalId: NotRequired[bool]

    # The name of the search result, including municipality name
    name: str

    # The name of the search result, without municipality name
    disassembledName: str

    # The coordinates of the search result.
    coord: CoordTuple

    # The type of the result.
    type: StopFinderResultType

    # The street name, if search result is type "street or singlehouse"
    streetName: NotRequired[str]

    # The house number, if search result is "singlehouse".
    buildingNumber: NotRequired[str]

    # Quality of the query matching.
    matchQuality: int

    # Products at this stop
    productClasses: NotRequired[List[ProductClass]]

    # Principality of the stop or stop area.
    parent: StopArea



from typing import List, Optional, cast

from ..models.deviations import Deviation, TransportMode
from .common import AsyncClient, UrlParams

__all__ = ("DeviationsClient",)


class DeviationsClient(AsyncClient):
    """
    client for SL Deviations API
    https://www.trafiklab.se/api/trafiklab-apis/sl/deviations/#/
    """

    @staticmethod
    def get_request_url_params(
        future: Optional[bool] = None,
        site: Optional[List[int]] = None,
        line: Optional[List[str]] = None,
        transport_authority: Optional[int] = None,
        transport_mode: Optional[List[TransportMode]] = None,
    ) -> UrlParams:
        """returns url and params to request deviations"""

        params: List[tuple[str, str]] = []
        if future is not None:
            params.append(("future", "true" if future else "false"))
        if site is not None:
            params.extend(("site", str(x)) for x in site)
        if line is not None:
            params.extend(("line", str(x)) for x in line)
        if transport_authority is not None:
            params.append(("transport_authority", str(transport_authority)))
        if transport_mode is not None:
            params.extend(("transport_mode", x.value) for x in transport_mode)

        return UrlParams("https://deviations.integration.sl.se/v1/messages", params)

    async def get_deviations(
        self,
        future: Optional[bool] = None,
        site: Optional[List[int]] = None,
        line: Optional[List[str]] = None,
        transport_authority: Optional[int] = None,
        transport_mode: Optional[List[TransportMode]] = None,
    ) -> List[Deviation]:

        args = self.get_request_url_params(
            future, site, line, transport_authority, transport_mode
        )
        response = await self._request_json(args)
        return cast(List[Deviation], response)



from typing import List, cast
from enum import IntFlag

from ..models.stops import StopFinderType
from .common import AsyncClient, ResponseFormatChanged, UrlParams


class Filter(IntFlag):
    STOPS = 2
    STREET = 4
    ADDRESS = 8
    POI = 32

class StopLookupClient(AsyncClient):
    """
    client for SL Stop Lookup API
    https://www.trafiklab.se/sv/api/our-apis/sl/journey-planner-2/
    """

    @staticmethod
    def _get_request_url_params(search_string: str, filter: Filter = Filter.STOPS) -> UrlParams:
        """returns url and params to request stops"""

        params = [
            ("name_sf", search_string),
            ("type_sf", "any"),
            ("any_obj_filter_sf", str(filter.value)),
        ]

        return UrlParams(
            "https://journeyplanner.integration.sl.se/v2/stop-finder",
            params,
        )

    async def _request_matches(self, args: UrlParams) -> List[StopFinderType]:
        response = await self._request_json(args)

        if (locations := response.get("locations")) is None:
            raise ResponseFormatChanged("'ResponseData' not found in response")

        locations = cast(List[StopFinderType], locations)
        return sorted(locations, key=lambda x: x["matchQuality"], reverse=True)

    async def get_stops(self, search_string: str) -> List[StopFinderType]:
        """
        Get stops by search string
        """

        args = self._get_request_url_params(search_string)
        return await self._request_matches(args)

    async def get_address(self, search_string: str) -> List[StopFinderType]:
        """
        Get street or address by search string
        """

        args = self._get_request_url_params(search_string, filter=Filter.STREET | Filter.ADDRESS)
        return await self._request_matches(args)



from typing import List, TypedDict, NotRequired

ValidityPeriod = TypedDict(
    "ValidityPeriod",
    {
        "from": str,
        "to": NotRequired[str],
    },
)
"""The period for which the object is valid"""


class Site(TypedDict):
    # Unique identifier of a site
    id: int

    # Global unique site identifier
    gid: int

    # The name of the site generaly known to the public
    name: str

    # The period for which the object is valid.
    valid: ValidityPeriod

    # Alias names that describes the same place but with a different name
    alias: NotRequired[List[str]]

    # An abbreviation for the site
    abbreviation: NotRequired[str]

    # Additional note related to the site
    note: NotRequired[str]

    # WGS84 latitude in decimal degrees
    lat: NotRequired[float]

    # WGS84 longitude in decimal degrees
    lon: NotRequired[float]



from enum import StrEnum
from typing import List, NotRequired, TypedDict

from .common import TransportMode


class DepartureState(StrEnum):
    NOT_EXPECTED = "NOTEXPECTED"
    NOT_CALLED = "NOTCALLED"
    EXPECTED = "EXPECTED"
    CANCELLED = "CANCELLED"
    INHIBITED = "INHIBITED"
    AT_STOP = "ATSTOP"
    BOARDING = "BOARDING"
    BOARDING_CLOSED = "BOARDINGCLOSED"
    DEPARTED = "DEPARTED"
    PASSED = "PASSED"
    MISSED = "MISSED"
    REPLACED = "REPLACED"
    ASSUMED_DEPARTED = "ASSUMEDDEPARTED"


class JourneyState(StrEnum):
    # Do not show departure at all. Some systems might instead indicate that this departure is available only if ordered
    NOT_EXPECTED = "NOTEXPECTED"

    # If a not expected dated vehicle journey is never run, it should at some point in time be considered as not run
    NOT_RUN = "NOTRUN"

    # Normally show target time for departure
    EXPECTED = "EXPECTED"

    # A symbol or text indicating that the vehicle journey is not yet in progress could be added depending on presentation system configuration
    ASSIGNED = "ASSIGNED"

    # Show departure as cancelled
    CANCELLED = "CANCELLED"

    # If the presentation system only shows vehicles that are in progress, do not show the departure
    SIGNED_ON = "SIGNEDON"

    # Normally show target time for departure. A symbol or text indicating that the vehicle journey is at origin, but not yet in progress could be added depending on presentation system configuration. If the presentation system only shows vehicles that are in progress, do not show the departure
    AT_ORIGIN = "ATORIGIN"

    # Present the current vehicle journey position, i.e. "has left station X Z minutes ago" or "currently at station Y". Systems that cannot present texts of that size should use a symbol or text indicating that the vehicle journey prediction is unreliable
    FAST_PROGRESS = "FASTPROGRESS"

    # Present the current vehicle journey position, i.e. "has left station X Z minutes ago" or "currently at station Y"
    NORMAL_PROGRESS = "NORMALPROGRESS"

    # Present the current vehicle journey position, i.e. "has left station X Z minutes ago" or "currently at station Y" and information that "traffic moves slowly". Systems that cannot present texts of that size should use a symbol or text indicating that the vehicle journey prediction is unreliable
    SLOW_PROGRESS = "SLOWPROGRESS"

    # Present the current vehicle journey position, i.e. "has left station X Z minutes ago" or "currently at station Y" and information that there is a "stop in traffic". Systems that cannot present texts of that size should use a symbol or text indicating that the vehicle journey prediction is unreliable
    NO_PROGRESS = "NOPROGRESS"

    # If the vehicle system detects that a vehicle is not following the expected route, it can change the state to off route
    OFF_ROUTE = "OFFROUTE"

    # If the vehicle finally reaches its destination, the vehicle journey receives the state completed, else it is aborted. Once in progress, a cancellation or sign off will be regarded as the monitored vehicle journey has been aborted. If an aborted dated vehicle journey is resumed again, PubTrans will create a new instance of a monitored vehicle journey
    ABORTED = "ABORTED"

    # If the vehicle finally reaches its destination, the vehicle journey receives the state completed, else it is aborted
    COMPLETED = "COMPLETED"

    # If an expected vehicle journey is not cancelled and never becomes in progress, it should at some point in time be considered as assumed completed
    ASSUMED_COMPLETED = "ASSUMEDCOMPLETED"


class JourneyPredictionState(StrEnum):
    NORMAL = "NORMAL"
    LOST_CONTACT = "LOSTCONTACT"
    UNRELIABLE = "UNRELIABLE"


class JourneyPassengerLevel(StrEnum):
    EMPTY = "EMPTY"
    SEATS_AVAILABLE = "SEATSAVAILABLE"
    STANDING_PASSENGERS = "STANDINGPASSENGERS"
    PASSENGERS_LEFT_BEHIND = "PASSENGERSLEFTBEHIND"
    UNKNOWN = "UNKNOWN"


class StopAreaType(StrEnum):
    BUS_TERMINAL = "BUSTERM"
    METRO_STATION = "METROSTN"
    TRAM_STATION = "TRAMSTN"
    RAILWAY_STATION = "RAILWSTN"
    SHIP_BERTH = "SHIPBER"
    FERRY_BERTH = "FERRYBER"
    AIRPORT = "AIRPORT"
    TAXI_TERMINAL = "TAXITERM"
    UNKNOWN = "UNKNOWN"


class DepartureJourney(TypedDict):
    id: int
    state: JourneyState
    prediction_state: NotRequired[JourneyPredictionState]
    passenger_level: NotRequired[JourneyPassengerLevel]


class StopAreaReference(TypedDict):
    id: int
    name: str
    sname: NotRequired[str]
    type: NotRequired[StopAreaType]


class StopPointReference(TypedDict):
    id: int
    name: NotRequired[str]
    designation: NotRequired[str]


class LineReference(TypedDict):
    id: int
    designation: NotRequired[str]
    transport_mode: NotRequired[TransportMode]
    group_of_lines: NotRequired[str]


class DepartureDeviation(TypedDict):
    importance_level: int
    consequence: str
    message: str


class Departure(TypedDict):
    direction: str
    direction_code: int
    state: DepartureState
    display: str
    journey: DepartureJourney
    stop_area: StopAreaReference
    stop_point: StopPointReference
    line: LineReference
    deviations: List[DepartureDeviation]
    scheduled: str
    expected: NotRequired[str]
    via: NotRequired[str]
    destination: NotRequired[str]


class DeviationScope(TypedDict):
    description: NotRequired[str]
    lines: NotRequired[List[LineReference]]
    stop_areas: NotRequired[List[StopAreaReference]]
    stop_points: NotRequired[List[StopPointReference]]


class StopDeviation(TypedDict):
    id: int
    importance_level: int
    message: str
    scope: DeviationScope


class SiteDepartureResponse(TypedDict):
    departures: List[Departure]
    stop_deviations: List[StopDeviation]



import logging
from functools import cached_property
from typing import Any, Dict, List, NamedTuple, Tuple, Union

import aiohttp

from .. import __version__

Params = Union[List[Tuple[str, str]], Dict[str, Any], None]

PARAM_KEY = "key"
SENSITIVE_PARAMS = {PARAM_KEY}


class UrlParams(NamedTuple):
    url: str
    params: Params


class AsyncClient:
    def __init__(self, session: aiohttp.ClientSession) -> None:
        self._session = session
        self.logger = logging.getLogger(self.__class__.__name__)

    @cached_property
    def user_agent(self):
        return f"{aiohttp.http.SERVER_SOFTWARE} trafiklab-sl/{__version__}"

    def _safe_params(self, params: Params) -> Params:
        if isinstance(params, dict):
            safe_params = params.copy()
        elif isinstance(params, list):
            safe_params = dict(params)
        else:
            safe_params = {}

        for key in SENSITIVE_PARAMS:
            if key in safe_params:
                safe_params[key] = "*****"

        return safe_params

    async def _request_json(self, args: UrlParams):
        self.logger.debug(
            f"Requesting {args.url} with params {self._safe_params(args.params)}"
        )
        response = await self._session.get(
            args.url,
            params=args.params,
            headers={
                "Content-Type": "application/json",
                "User-Agent": self.user_agent,
            },
        )
        response.raise_for_status()
        json = await response.json()
        return json


class ClientException(Exception):
    """Base class for client exceptions"""


class ResponseFormatChanged(ClientException):
    """Raised when the response format has changed"""


class OperationFailed(ClientException):
    """Raised when an operation failed with known error code"""

    def __init__(self, errors: List[str]):
        self.errors = errors
        super().__init__(f"Operation failed with errors: {', '.join(errors)}")



import re
from datetime import datetime
from zoneinfo import ZoneInfo

SL_TZ = ZoneInfo("Europe/Stockholm")
ISO_MSEC_PART = re.compile(r"\.(\d{1,3})\+?")


def from_sl_dt(dt_str: str) -> datetime:
    """
    Convert a datetime string from SL format to a timezone-aware datetime object.

    The SL datetime format is expected to be in ISO 8601 format, e.g., "2023-10-01T12:00:00".
    """

    dt = datetime.fromisoformat(dt_str)
    return dt.astimezone(SL_TZ)


def global_id_to_site_id(global_id: str) -> int:
    """
    Convert a global ID to a transport site ID.

    Function assumes the global ID is in the format "909100100xxxxxxx",
    """

    if len(global_id) != 16:
        raise ValueError(
            f"Invalid global ID length: {len(global_id)}. Expected 16 characters."
        )

    prefix = "909100100"
    if not global_id.startswith(prefix):
        raise ValueError(
            f"Invalid global ID prefix: {global_id[:9]}. Expected '{prefix}'."
        )

    return int(global_id.removeprefix(prefix))



from typing import List

from tsl.models.journey import DwellTime, Journey, Language, RouteType, SearchLeg

from .common import AsyncClient, OperationFailed, ResponseFormatChanged, UrlParams


class JourneyPlannerClient(AsyncClient):
    """
    client for SL Journey Planner v2 API
    https://www.trafiklab.se/sv/api/our-apis/sl/journey-planner-2/
    """

    @staticmethod
    def build_request_params(
        origin: SearchLeg,
        destination: SearchLeg,
        calc_number_of_trips: int = 1,
        via: SearchLeg | None = None,
        not_via: SearchLeg | None = None,
        dwell_time: DwellTime | None = None,
        language: Language | None = None,
        max_changes: int | None = None,
        route_type: RouteType | None = None,
        include_coordinates: bool | None = None,
    ) -> UrlParams:
        """returns url and params to request journey planner API"""

        if calc_number_of_trips > 3 or calc_number_of_trips < 1:
            raise ValueError("calc_number_of_trips must be between 1 and 3")

        params = [
            ("type_origin", origin.type),
            ("name_origin", origin.value),
            ("type_destination", destination.type),
            ("name_destination", destination.value),
            ("calc_number_of_trips", calc_number_of_trips)
        ]

        if via is not None:
            params.append(("type_via", via.type))
            params.append(("name_via", via.value))

        if not_via is not None:
            params.append(("type_not_via", not_via.type))
            params.append(("name_not_via", not_via.value))

        if dwell_time is not None:
            params.append(("dwell_time", dwell_time.to_hh_mm()))

        if language is not None:
            params.append(("language", language))

        if max_changes is not None:
            params.append(("max_changes", max_changes))

        if route_type is not None:
            params.append(("route_type", route_type.value))

        if include_coordinates is not None:
            params.append(("gen_c", 'true' if include_coordinates else 'false'))

        return UrlParams(
            "https://journeyplanner.integration.sl.se/v2/trips",
            params,
        )

    async def search_trip(self, params: UrlParams) -> List[Journey]:
        response = await self._request_json(params)

        if (journeys := response.get("journeys")) is None:
            if messages := response.get("systemMessages"):
                filtered_errors = [
                    x["text"]
                    for x in messages
                    if x["type"] == "error" and x["code"] != -8010
                ]
                raise OperationFailed(filtered_errors)

            raise ResponseFormatChanged(
                "Response format has changed, expected 'journeys' or `systemMessages` keys in response"
            )

        return journeys
